"""Plotting functionality."""

# Not part of dependencies because not used by default, only for debugging.
import matplotlib.pyplot  # noqa: ICN001

plot = matplotlib.pyplot.plot
show = matplotlib.pyplot.show
subplots = matplotlib.pyplot.subplots
